// Copyright 2021 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

// This file encapsulates some of the odd characteristics of the
// LOONGARCH64 (LOONGARCH64) instruction set, to minimize its interaction
// with the core of the assembler.

package arch

import (
	"cmd/internal/obj"
	"cmd/internal/obj/loongarch64"
)

func jumpLOONGARCH64(word string) bool {
	switch word {
	case "BEQ", "BFPF", "BFPT", "BLTZ", "BGEZ", "BLEZ", "BGTZ", "BLT", "BLTU", "JIRL", "BNE", "BGE", "BGEU", "JMP", "JAL", "CALL":
		return true
	}
	return false
}

// IsLOONGARCH64CMP reports whether the op (as defined by an loongarch64.A* constant) is
// one of the CMP instructions that require special handling.
func IsLOONGARCH64CMP(op obj.As) bool {
	switch op {
	case loongarch64.ACMPEQF, loongarch64.ACMPEQD, loongarch64.ACMPGEF, loongarch64.ACMPGED,
		loongarch64.ACMPGTF, loongarch64.ACMPGTD:
		return true
	}
	return false
}

// IsLOONGARCH64MUL reports whether the op (as defined by an loongarch64.A* constant) is
// one of the MUL/DIV/REM/MADD/MSUB instructions that require special handling.
func IsLOONGARCH64MUL(op obj.As) bool {
	switch op {
	case loongarch64.AMUL, loongarch64.AMULU, loongarch64.AMULV, loongarch64.AMULVU,
		loongarch64.ADIV, loongarch64.ADIVU, loongarch64.ADIVV, loongarch64.ADIVVU,
		loongarch64.AREM, loongarch64.AREMU, loongarch64.AREMV, loongarch64.AREMVU:
		return true
	}
	return false
}

func loongarch64RegisterNumber(name string, n int16) (int16, bool) {
	switch name {
	case "F":
		if 0 <= n && n <= 31 {
			return loongarch64.REG_F0 + n, true
		}
	case "FCSR":
		if 0 <= n && n <= 31 {
			return loongarch64.REG_FCSR0 + n, true
		}
	case "FCC":
		if 0 <= n && n <= 31 {
			return loongarch64.REG_FCC0 + n, true
		}
	case "R":
		if 0 <= n && n <= 31 {
			return loongarch64.REG_R0 + n, true
		}
		//MSA
		//	case "W":
		//		if 0 <= n && n <= 31 {
		//			return loongarch64.REG_W0 + n, true
		//		}
	}
	return 0, false
}
